@@ -46,7 +46,11 @@
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
 #include <netinet/udp.h>
+#include <linux/ip.h>
+#include <linux/icmp.h>
 #include <linux/netdevice.h>
+#include <linux/etherdevice.h>
+#include <linux/skbuff.h>
 
 #include "sf_custom.h"
 #include "sf_filter.h"
@@ -60,6 +64,10 @@
 
 #undef LOG_DEBUG
 
+#ifndef IP_OFFSET
+#define IP_OFFSET  0x1FFF
+#endif
+
 /*
  * Public variables 
  */
@@ -72,7 +80,8 @@ int fddev = 0;                          /* device file handle */
  * Static variables
  */
 
-static FILE *fdlog         = NULL;      /* file pointer to above log file */
+static FILE *fdlog         = NULL;      /* file pointer to log file */
+static FILE *fdreport      = NULL;      /* file pointer to report file */
 static char *last_log      = NULL;	/* last log message */ 
 static char *last_msg      = NULL;      /* last user message */
 static char num_last_log   = 0;		/* number of repeated messages */
@@ -88,17 +97,9 @@ static struct timeout *event_queue = NULL; /* events to be processed */
 static char tmp[255];                   /* used for sprintf() */
 static struct sf_fw *dynrules = NULL;   /* chain containing all dynamic rules */
 
-
-static char *action[]   = {"reject(port unr)", 
-                           "reject(protocol unr)",
-                           "reject(host unr)",
-                           "reject(net unr)",
-                           "block",
-                           "accept"};
-
 /*
- * Default action for IP spoofing notification -- this case must be
- * treated specially, because spoofing is an implicit rule.
+ * Default actions for IP spoofing notification and oversized packets 
+ * -- these are implicit rules
  */
 
 static struct notification notification_spoof = {
@@ -106,6 +107,7 @@ static struct notification notification_spoof = {
   NULL,  /* n_next        */
   0,     /* n_syslog      */
   0,     /* n_spy         */
+  0,     /* n_report      */
   0,     /* n_relevel     */
   {0},   /* n_relevel_val */
   SPOOF_MESSAGE,
@@ -125,6 +127,31 @@ struct sf_fw rule_spoof = {
 };
 
 
+static struct notification notification_oversized = {
+  0,     /* n_level       */
+  NULL,  /* n_next        */
+  0,     /* n_syslog      */
+  0,     /* n_spy         */
+  0,     /* n_report      */
+  0,     /* n_relevel     */
+  {0},   /* n_relevel_val */
+  OVERSIZED_MESSAGE,
+  NULL,
+  NULL,
+  NULL,
+  NULL 
+};
+
+struct sf_fw rule_oversized = {
+  SF_FW_MAGIC,  /* magic   */
+  NULL,         /* fw_next */
+  0,            /* fw_line */
+  {0},          /* rule_id */
+  {&notification_oversized},
+  SF_RULE_OVERSIZED
+};
+
+
 /*
  * Definitions for the event manager: event queue and signal handlers
  */
@@ -156,6 +183,7 @@ void prepare_for_exec(void)
   close(fdlock);
   close(fddev); 
   fclose(fdlog);
+  fclose(fdreport);
  
   hasclosed = 1;
 }
@@ -179,7 +207,7 @@ static void get_time(void)
 
 
 /*
- * Connect to the operating system's log daemon and set up our log file
+ * Connect to the operating system's log daemon and set up our log files
  */
 
 static void connect_logs(char *config)
@@ -190,13 +218,18 @@ static void connect_logs(char *config)
     perror("Cannot append to log file");
     exit(1);
   }
+  if ((fdreport = fopen(REPORT, "a")) == NULL) {
+    perror("Cannot append to report file");
+    exit(1);
+  }
 
-  /* The log file must not be fully buffered... We want to be able to read
-   * it even if the system crashes.
+  /* The log files must not be fully buffered... We want to be able to read
+   * them even if the system crashes.
    */
-  setvbuf(fdlog, NULL, _IOLBF, 0);
-  setvbuf(stderr, NULL, _IOLBF, 0);
-
+  setvbuf(fdlog,    NULL, _IOLBF, 0);
+  setvbuf(fdreport, NULL, _IOLBF, 0);
+  setvbuf(stderr,   NULL, _IOLBF, 0);
+  
   fprintf(fdlog, "%s daemon started, config: %s\n", lstime, config);
   
   /* all's well that ends well */
@@ -307,6 +340,7 @@ static void crit_error(char *msg)
   log_error(msg);
 
   fclose(fdlog);
+  fclose(fdreport);
 
   /* Well, maybe one shoudn't clutter the console... */
   fprintf(stderr, "\007Firewall daemon -- Fatal error: %s!\n", msg);
@@ -441,6 +475,13 @@ static int fetch_op_value(int optype, int operand, struct iphdr *ip)
     case OPERAND_VAR_DESTHOST:   
          return fetch_var_value(operand, ip->daddr);
     break;
+    case OPERAND_DESTPORT:
+	 if ((ip->protocol != IPPROTO_UDP) && (ip->protocol != IPPROTO_TCP))
+	   return 0;
+	 else 
+	   return ntohs(((struct tcphdr *)
+                  ((unsigned long *)ip+ip->ihl))->th_dport);
+    break;
   }
 
   return 0; /* default and OPERAND_NOT_USED */
@@ -656,11 +697,12 @@ static void remove_rule(int id, int tid)
 /*
  * Execute the statements in a notifcation body.
  * This routine will be called recursively, because we have `if' 
- * statements.
+ * and call statements.
  */
 
 static void execute_notify(struct sf_fw *rule, char *message,
-                           struct notification *note, struct iphdr *ip)
+                           struct notification *note, 
+			   struct iphdr *ip)
 {
   union {
     struct exec_chain   *_exec;
@@ -669,7 +711,7 @@ static void execute_notify(struct sf_fw *rule, char *message,
   } p;
 
 
-  /* personal log message? */
+  /* customized log message? */
   if (note->n_message != NULL) {
     if ((last_msg == NULL) || (strcmp(last_msg, note->n_message) != 0)) {
       if (last_msg != NULL)
@@ -679,14 +721,19 @@ static void execute_notify(struct sf_fw *rule, char *message,
     }
   }
 
-
   /* syslog the message? */
-  if (note->n_syslog)  {
+  if (note->n_syslog) {
     syslog(LOG_AUTHPRIV|LOG_CRIT, message);
     if (note->n_message != NULL)
       syslog(LOG_AUTHPRIV|LOG_CRIT, note->n_message);
   }
 
+  /* include message in report? */
+  if (note->n_report) {
+    flogf(fdreport, lstime, message);
+    if (note->n_message != NULL)
+      flogf(fdreport, lstime, note->n_message);
+  }
 
   /* change log level? */
   if (note->n_relevel)
@@ -794,7 +841,10 @@ static void execute_notify(struct sf_fw *rule, char *message,
       variables[p._li->variable].timeout =
         (p._li->timeout == 0) ? INT_MAX : ltime+p._li->timeout;
     }
-
+    else if (p._li->statement_type == CALL_STATEMENT) {
+      if (p._li->then)
+        execute_notify(rule, message, p._li->then, ip);
+    }
     else { /* if statement */
       int cond;
 
@@ -811,7 +861,7 @@ static void execute_notify(struct sf_fw *rule, char *message,
       }
 
       if (cond && p._li->then)
-        execute_notify(rule, message, p._li->then, ip); /* recursive call */
+        execute_notify(rule, message, p._li->then, ip); /* recursive */
     }
   }
 
@@ -881,15 +931,17 @@ static void execute_notify(struct sf_fw *rule, char *message,
 
 static void packet_loop(void)
 {
-  struct iphdr  *ip;
-  struct tcphdr *tcp;
+  struct iphdr   *ip;
+  struct tcphdr  *tcp;
+  struct icmphdr *icmp;
   char cs[20], cd[20], rule_desc[20], tmp[255];
-  char message[255], source_port[85], target_port[85], protocol_desc[15];
+  char message[255], source_port[85], target_port[85], protocol_desc[50];
   int i, protocol_index;
   struct sf_proc log_data;
   struct sf_fw *rule;
   struct statfs statbuf;
   long free_blocks;
+  char *action;
 
 
   while (1) {
@@ -947,12 +999,16 @@ static void packet_loop(void)
       /* handle implicit rules */
       if (((int) rule == RULE_SPOOF_RECV) || ((int) rule == RULE_SPOOF_XMIT))
         rule = &rule_spoof;
+      else
+	if ((int) rule == RULE_OVERSIZED)
+	  rule = &rule_oversized;
 
       if ((rule == NULL) || (rule->level.ptr == NULL))
         continue; /* there might be an obsolete entry in the buffer */
 
       ip   = (struct iphdr *)  log_data.data[i].ip_packet;
       tcp  = (struct tcphdr *) ((unsigned long *)ip + ip->ihl);
+      icmp = (struct icmphdr *) tcp;
 
       /* Compose the mandatory log entry.
          Log entries consist of:
@@ -967,34 +1023,48 @@ static void packet_loop(void)
       *source_port =
       *target_port = 0;
 
-      if (ip->protocol < MAX_PROTOCOL) {
+      if (ip->protocol <= MAX_PROTOCOL) {
         protocol_index = ip->protocol;
-        if ((ip->protocol == IPPROTO_UDP) ||
-            (ip->protocol == IPPROTO_TCP)) {
-          struct servent *serv_ent;
-          char proto[4];
+	if ((ntohs(ip->frag_off)&IP_OFFSET) == 0) { /* not a fragment */
+          if ((ip->protocol == IPPROTO_UDP) ||
+              (ip->protocol == IPPROTO_TCP)) {
+            struct servent *serv_ent;
+            char proto[4];
 
-          if (ip->protocol == IPPROTO_UDP) 
-            strcpy(proto, "udp");
-          else
-            strcpy(proto, "tcp");
+            if (ip->protocol == IPPROTO_UDP) 
+              strcpy(proto, "udp");
+            else
+              strcpy(proto, "tcp");
 
-          if ((serv_ent = getservbyport(tcp->th_sport, proto)) != NULL)
-            sprintf(source_port, ":%s", serv_ent->s_name);
-          else
-            sprintf(source_port, ":%i", ntohs(tcp->th_sport));
+            if ((ip->protocol == IPPROTO_UDP) && ((serv_ent = getservbyport(tcp->th_sport, proto)) != NULL))
+              sprintf(source_port, ":%s", serv_ent->s_name);
+            else
+	      sprintf(source_port, ":%i", ntohs(tcp->th_sport));
 
-          if ((serv_ent = getservbyport(tcp->th_dport, proto)) != NULL)
-            sprintf(target_port, ":%s", serv_ent->s_name);
-          else
-            sprintf(target_port, ":%i", ntohs(tcp->th_dport));
-        }
+            if ((serv_ent = getservbyport(tcp->th_dport, proto)) != NULL)
+              sprintf(target_port, ":%s", serv_ent->s_name);
+            else
+              sprintf(target_port, ":%i", ntohs(tcp->th_dport));
+          }
+          else {
+            if (ip->protocol == IPPROTO_ICMP) {
+              if (icmp->type <= MAX_ICMPTYPE)
+	        sprintf(protocol_desc, " %s", icmptype[icmp->type]);
+	      else
+                sprintf(protocol_desc, " type %i", icmp->type);
+            }
+	  }
+        } /* not a fragment */
       }
       else {
         protocol_index = MAX_PROTOCOL+1;
         sprintf(protocol_desc, "%i", ip->protocol);
       }
 
+      if ((ntohs(ip->frag_off)&IP_OFFSET) != 0) { /* fragment */
+        sprintf(protocol_desc, " frag.");
+      }
+
       strcpy(cs, inet_ntoa(*((struct in_addr *) &(ip->saddr))));
       strcpy(cd, inet_ntoa(*((struct in_addr *) &(ip->daddr))));
 
@@ -1002,12 +1072,25 @@ static void packet_loop(void)
         sprintf(rule_desc, "(s %i)", rule->fw_line);
       else if (rule->timeout == SF_RULE_SPOOFING)
         sprintf(rule_desc, "SPOOF: %s", log_data.data[i].devname);
+      else if (rule->timeout == SF_RULE_OVERSIZED)
+        sprintf(rule_desc, "XSIZE: %s", log_data.data[i].devname);
       else 
         sprintf(rule_desc, "(d %i)", rule->fw_line);
 
+      switch (log_data.data[i].action) {
+        case FW_ACCEPT:     action = "accept"; break;
+        case FW_BLOCK:      action = "block"; break;
+        case FW_REJECT:     action = "reject (host unr)"; break;
+	case SF_RC_TREJECT: action = "reject (best/tcp)"; break;
+        case SF_RC_ECHO:    action = "reject (echo reply)"; break;
+	case SF_RC_RNET:    action = "reject (net unr)"; break;
+        case SF_RC_RPROTO:  action = "reject (protocol unr)"; break;
+        case SF_RC_RPORT:   action = "reject (port unr)"; break;
+        default:            action = "???"; break;
+      }
+
       sprintf(message, "%s %s %s%s %s%s->%s%s",
-              rule_desc,            
-              action[log_data.data[i].action-SF_RC_MIN],
+              rule_desc, action,
               protocol[protocol_index], protocol_desc,
               cs, source_port,
               cd, target_port);
