@@ -27,17 +27,8 @@
  *
  */
 
-#include <linux/config.h>
-
-#ifdef MODULE
-#include <linux/module.h>
-#include <linux/version.h>
-#else
-#define MOD_INC_USE_COUNT
-#define MOD_DEC_USE_COUNT
-#endif
-
 #include <stdio.h>
+#include <linux/config.h>
 #include <linux/kernel.h>
 #include <linux/sched.h>
 #include <linux/string.h>
@@ -52,14 +43,19 @@
 #include <linux/ip.h>
 #include <linux/malloc.h>
 #include <linux/timer.h>
-
-#include <linux/sf_kernel.h>
+#include <net/ip.h>
+#include <linux/igmp.h>
+#include "sf_kernel.h"
 #include "sf_rc.h"
 #include "sf_global.h"
 #include "sf_filter.h"
 #include "sf_control.h"
+#include "sf_tcp.h"
+#include "sf_icmp.h"
 
-#include "/usr/src/linux/net/inet/ip.h"
+/* loopback is the first device in the dev chain */
+#define LOOPBACK_MASK (dev_base->pa_mask)
+#define LOOPBACK_ADDR (dev_base->pa_addr) 
 
 /* firewall enabled */
 int sf_fw_enabled = 0;
@@ -170,7 +166,7 @@ struct sf_fragment {
 	__u32 dst;			/* destination address */
 	struct timer_list timer; 	/* expiration of datagram */
 };
-#define SF_FRAG_HASH_SIZE	199	/* must be a prime number */
+#define SF_FRAG_HASH_SIZE	499	/* must be a prime number */
 struct sf_fragment *sf_frag[SF_FRAG_HASH_SIZE];
 #define SF_FRAG_KEY(a)	( (a) % SF_FRAG_HASH_SIZE )
 
@@ -211,13 +207,13 @@ static void sf_frag_insert(__u16 id, __u32 saddr, __u32 daddr)
     frag->id  = id;
     frag->src = saddr;
     frag->dst = daddr;
-#ifdef SF_DEBUG
+#ifdef SF_DEBUG_FRAG
     printk("sf_frag_insert: id = %u\n",ntohs(frag->id));
 #endif
 
     /* initialize timer */
     init_timer(&frag->timer);
-    frag->timer.expires = IP_FRAG_TIME; /* same time as in reassembly */
+    frag->timer.expires = jiffies + IP_FRAG_TIME; /* same time as in reassembly */
     frag->timer.data = (unsigned long) frag;
     frag->timer.function = sf_frag_expire;
     add_timer(&frag->timer);
@@ -254,12 +250,12 @@ struct sf_tcp_connection {
 	struct sf_tcp_connection *prev, *next;
 	__u32 client_addr, server_addr;	
 	__u16 client_port, server_port;
-	int state;			/* simpified state of tcp connection */
+	int state;			/* simplified state of tcp connection */
 	struct timer_list timer;
 	int timer_set;			/* one, if a timer is set */
 };
 
-#define SF_TCP_HASH_SIZE	199	/* must be a prime number */
+#define SF_TCP_HASH_SIZE	499	/* must be a prime number */
 struct sf_tcp_connection *sf_tcp[SF_TCP_HASH_SIZE];
 #define SF_TCP_KEY(a)  ( (a) % SF_TCP_HASH_SIZE )
 
@@ -413,6 +409,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 	struct icmphdr *icmp=(struct icmphdr *)((unsigned long *)ip+ip->ihl);
 	struct igmphdr *igmp=(struct igmphdr *)((unsigned long *)ip+ip->ihl);
 	struct sf_tcp_connection *sftcp = NULL;
+	int returncode = FW_BLOCK;
 
 #ifdef SF_DEBUG
   if (ip->protocol == IPPROTO_ICMP) {
@@ -430,20 +427,20 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 
 	/* let packet pass if firewall functions are disabled */
 	if (sf_fw_enabled <= 0) 
-	  return SF_RC_BLOCK;
+	  return FW_BLOCK;
 
         /* check if packet is long enough to hold IP header */
-        if ( ntohs(ip->tot_len) < (4 * ip->ihl) ) return SF_RC_BLOCK;
-        if ( ip->ihl < 5 ) return SF_RC_BLOCK;
+        if ( ntohs(ip->tot_len) < (4 * ip->ihl) ) return FW_BLOCK;
+        if ( ip->ihl < 5 ) return FW_BLOCK;
 
-	/* if the device is loopback, source and dest addr. must be loopback addr. */
+	/* if the device Is loopback, source and dest addr. must be loopback addr. */
 	if ( !(rif->flags & IFF_LOOPBACK)
-	  && ( ((ip->saddr & loopback_dev.pa_mask) == 
-			(loopback_dev.pa_addr & loopback_dev.pa_mask) )
-	    || ((ip->daddr & loopback_dev.pa_mask) == 
-			(loopback_dev.pa_addr & loopback_dev.pa_mask) ) ) )
+	  && ( ((ip->saddr & LOOPBACK_MASK) == 
+			(LOOPBACK_ADDR & LOOPBACK_MASK) )
+	    || ((ip->daddr & LOOPBACK_MASK) == 
+			(LOOPBACK_ADDR & LOOPBACK_MASK) ) ) )
             { /* packet not allowed */
-              rule->fw_rc = SF_RC_BLOCK;
+	      rule->fw_rc = FW_BLOCK;
               if (opt == SF_STATE_RECEIVE)
                 { rule->rule_id.num = RULE_SPOOF_RECV; }
               else
@@ -453,7 +450,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 
 	/* check local packets only once */
 	if ( (opt == SF_STATE_RECEIVE) && (ip->saddr == ip->daddr) )
-	  return SF_RC_ACCEPT;
+	  return FW_ACCEPT;
 
 	/* check packet */
 
@@ -484,7 +481,15 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		&& !(IN_CLASSD(ntohl(ip->daddr)))
 		&& (ifinside != sf_inside(ip->daddr)) ) ) )
 	    { /* packet not allowed */
-	      rule->fw_rc = SF_RC_BLOCK;
+	      rule->fw_rc = FW_BLOCK;
+
+#if 0
+	      /* (bootp) - drop packets from/to 0.0.0.0/255.255.255.255 */
+	      if ((ip->saddr == 0) || (ip->daddr == 0) ||
+		  (ip->saddr == 0xffffffff) || (ip->daddr == 0xffffffff))
+		return FW_BLOCK;
+#endif
+
 	      if (opt == SF_STATE_RECEIVE)
 		{ rule->rule_id.num = RULE_SPOOF_RECV; }
 	      else
@@ -495,13 +500,29 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 
 	/* If SF_STATE_FORWARD, the packet has already been examined */
 	/* at receive time. So skip the examination now. */
-	if (opt == SF_STATE_FORWARD) return SF_RC_ACCEPT;
+	if (opt == SF_STATE_FORWARD) return FW_ACCEPT;
 
 	/* check fragments */
 	if (ntohs(ip->frag_off) & IP_OFFSET) /* fragment offset != 0 */
 	  { /* this is an ip fragment but not the first one of the packet */
 	    struct sf_fragment *frag = sf_frag[SF_FRAG_KEY(ip->id)];
 
+#ifdef SF_DEBUG_FRAG
+	    printk("sf_check_packet: frag_off %u  tot_len %u\n",
+		   (ntohs(ip->frag_off)&IP_OFFSET)<<3,ntohs(ip->tot_len));
+#endif
+
+	    if ( ((ntohs(ip->frag_off)&IP_OFFSET) << 3) + 
+                 (unsigned int)ntohs(ip->tot_len) > 65535 ) 
+	      {
+#ifdef SF_DEBUG_FRAG
+	        printk("sf_check_packet: oversized packet detected!\n");
+#endif
+		rule->fw_rc = FW_BLOCK;
+                rule->rule_id.num = RULE_OVERSIZED;
+                goto informlog; /* skip all further examination but inform daemon */
+	      }
+
 	    while (frag != NULL)
 	      { if ( (ip->id==frag->id) && (ip->saddr==frag->src) && (ip->daddr==frag->dst) )
                   { /* fragment belongs to an allowed packet */
@@ -509,16 +530,16 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		    /* fragments need not arrive in order, thus there may still */
 		    /* be fragments missing when the last fragment arrives. */
 		    del_timer(&frag->timer);
-		    frag->timer.expires = IP_FRAG_TIME; /* same time as in reassembly */
+		    frag->timer.expires = jiffies + IP_FRAG_TIME; /* same time as in reassembly */
 		    frag->timer.data = (unsigned long) frag;
 		    frag->timer.function = sf_frag_expire;
 		    add_timer(&frag->timer);
-		    return SF_RC_ACCEPT;
+		    return FW_ACCEPT;
 		  }
 		frag = frag->next;
 	      }
 	      /* fragment must not pass */
-	      return SF_RC_BLOCK;
+	      return FW_BLOCK;
 	  }
 
         /* Now we know that the packet is not a subsequent fragment.
@@ -526,13 +547,13 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
            entire UDP, ICMP, IGMP (or TCP, see below) header */
         if (ip->protocol == IPPROTO_UDP)
           if (ntohs(ip->tot_len) < (4 * ip->ihl + sizeof(struct udphdr)))
-            return SF_RC_BLOCK;
+            return FW_BLOCK;
         if (ip->protocol == IPPROTO_ICMP)
           if (ntohs(ip->tot_len) < (4 * ip->ihl + sizeof(struct icmphdr)))
-            return SF_RC_BLOCK;
+            return FW_BLOCK;
         if (ip->protocol == IPPROTO_IGMP)
           if (ntohs(ip->tot_len) < (4 * ip->ihl + sizeof(struct igmphdr)))
-            return SF_RC_BLOCK;
+            return FW_BLOCK;
 
 	/* check tcp */
 	if (ip->protocol == IPPROTO_TCP) 
@@ -542,7 +563,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 
             /* check if packet is long enough to contain the TCP header */
             if (ntohs(ip->tot_len) < (4 * ip->ihl + sizeof(struct tcphdr)))
-              return SF_RC_BLOCK; 
+              return FW_BLOCK; 
 
             key = SF_TCP_KEY(ip->saddr + ip->daddr + (__u32)tcp->source + (__u32)tcp->dest);
 
@@ -588,9 +609,9 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 	    if (sftcp == NULL) /* connection not in hash queue */
 	      { if (tcp->syn && !tcp->ack) goto checkrules; /* new connection */
 #ifdef SF_DEBUG
-  printk("connection not in hash queue\n");
+                printk("connection not in hash queue\n");
 #endif
-	        return SF_RC_BLOCK; /* block packet */
+	        return FW_BLOCK; /* block packet */
 	      }
 
 	    /* connection found in hash queue */
@@ -650,16 +671,16 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 
 	      }
 
-	    if ( (sftcp->state >= SF_TCP_ESTABLISHED) && !tcp->fin && !tcp->rst )
+	    if ( (sftcp->state >= SF_TCP_ESTABLISHED) && !tcp->fin && !tcp->rst && !tcp->syn)
 	      { SF_CHKFRAG; /* check if packet is fragmented */
-	        return SF_RC_ACCEPT; /* packet is allowed */ 
+	        return FW_ACCEPT; /* packet is allowed */ 
 	      }
 	    if ( tcp->rst )
 	      { if (sftcp->timer_set) del_timer(&sftcp->timer);
 		sf_tcp_delete(sftcp); /* RST aborts an connection immediately */
 
 		SF_CHKFRAG; /* check if packet is fragmented */
-		return SF_RC_ACCEPT;
+		return FW_ACCEPT;
               }
 	    if ( tcp->syn )
 	      { if ( sftcp->state == SF_TCP_TERMINATED )
@@ -668,9 +689,17 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		    /* if the TCP protocol is implemented correctly. But if  */
 		    /* it happens nevertheless, we must not allow the        */
 		    /* connection without checking the rules.                */
+#ifdef SF_DEBUG_TCP
+		    printk("sf_check_packet: SYN && SF_TCP_TERMINATED\n");
+#endif
 		    if (sftcp->timer_set) del_timer(&sftcp->timer);
 		    sf_tcp_delete(sftcp);
-		    return SF_RC_BLOCK; /* The sender is going to retry... */
+		    return FW_BLOCK; /* The sender is going to retry... */
+		  }
+		if ( sftcp->state >= SF_TCP_ESTABLISHED )
+                  { /* This may be a resent SYN packet */
+		    SF_CHKFRAG; /* check if packet is fragmented */
+		    return FW_ACCEPT;
 		  }
 		if ( !reverse ) 
 		  { sftcp->state = SF_TCP_CLIENT_SYN;
@@ -678,7 +707,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		    printk("sf_check_packet: tcp client sent SYN\n");
 #endif
 		    if (sftcp->timer_set) del_timer(&sftcp->timer);
-		    sftcp->timer.expires = SF_TCP_SYN_TO;
+		    sftcp->timer.expires = jiffies + SF_TCP_SYN_TO;
 		    sftcp->timer.data = (unsigned long) sftcp;
 		    sftcp->timer.function = sf_tcp_expire;
 		    sftcp->timer_set = 1;
@@ -696,7 +725,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		      } 
 		  }
 		SF_CHKFRAG; /* check if packet is fragmented */
-		return SF_RC_ACCEPT;
+		return FW_ACCEPT;
 	      }
 	    if ( tcp->fin )
 	      { if ( reverse )
@@ -705,7 +734,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 		  { sftcp->state |= SF_TCP_CLIENT_FIN; }
 		if ( sftcp->state == SF_TCP_TERMINATED )
 		  { if (sftcp->timer_set) del_timer(&sftcp->timer); 
-		    sftcp->timer.expires = SF_TCP_FIN_WAIT;
+		    sftcp->timer.expires = jiffies + SF_TCP_FIN_WAIT;
 		    sftcp->timer.data = (unsigned long) sftcp;
 		    sftcp->timer.function = sf_tcp_expire;
 		    sftcp->timer_set = 1;
@@ -715,11 +744,32 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 #endif
 		  }
 		SF_CHKFRAG; /* check if packet is fragmented */
-		return SF_RC_ACCEPT;
+		return FW_ACCEPT;
 	      }
 
-	    printk("sf_check_packet: checking tcp connections: THIS SHOULD NEVER HAPPEN!\n");
-	    return SF_RC_BLOCK;
+	    /* Only packets with !SYN, !FIN, !RST that belong to connections
+               with state < SF_TCP_ESTABLISHED can arrive here. */
+
+            /* This should never happen - but it does happen, e.g. when
+               one side thinks the connection is still established and
+               the other side tries to establish a new connection to the
+               same port numbers. This can also happen if the connection
+               has died on one end only (e.g. rebooted PC). There's no
+               need to accept these strange packets, since the other
+               side of the connection will not accept them.
+               
+               Maybe we should send a RST packet here? */
+
+#ifdef SF_DEBUG_TCP
+            printk("sf_check_packet: TCP connection in strange state:\n");
+            if (reverse)  printk("s->c "); else printk("c->s ");
+            if (tcp->syn) printk("SYN ");
+            if (tcp->fin) printk("FIN ");
+            if (tcp->ack) printk("ACK ");
+            if (tcp->rst) printk("RST ");
+            printk("\n");
+#endif
+	    return FW_BLOCK;
 
 	  } /* check tcp */
 
@@ -933,9 +983,9 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 	    }
 	}
 
-	if (rule == NULL) return SF_RC_BLOCK; /* packet matches no rule */
+	if (rule == NULL) return FW_BLOCK; /* packet matches no rule */
 
-	if (rule->fw_rc == SF_RC_ACCEPT)
+	if (rule->fw_rc == FW_ACCEPT)
 	  { /* if a tcp packet arrives here, it is initiating a new, legal connection */
 	    if (ip->protocol == IPPROTO_TCP) 
 	      { sftcp = sf_tcp_insert(ip->saddr, ip->daddr, tcp->source, tcp->dest);
@@ -944,7 +994,7 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 #ifdef SF_DEBUG
 		    printk("sf_check_packet: tcp client sent SYN\n");
 #endif
-		    sftcp->timer.expires = SF_TCP_SYN_TO;
+		    sftcp->timer.expires = jiffies + SF_TCP_SYN_TO;
 		    sftcp->timer.data = (unsigned long) sftcp;
 		    sftcp->timer.function = sf_tcp_expire;
 		    add_timer(&sftcp->timer); 
@@ -954,19 +1004,69 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 	    SF_CHKFRAG; /* check if packet is fragmented */
 	  }
 
-	if ( !(rule->fw_flags & SF_FW_LOG) ) return rule->fw_rc; /* no notification */
+	returncode = rule->fw_rc;
 
-      informlog:
+	if (rule->fw_rc == SF_RC_TREJECT) /* TCP reset / best reject*/
+	  {
+	    if (opt == SF_STATE_TRANSMIT)
+	      returncode = FW_REJECT;
+	    else 
+              { /* choose best method */
+	        if (ip->protocol == IPPROTO_UDP)
+	          { /* send ICMP port unreach. */
+                    send_icmp_error(rif, ip, ICMP_DEST_UNREACH, 
+                                    ICMP_PORT_UNREACH);
+		    returncode = FW_BLOCK;
+		  }
+		else if (ip->protocol == IPPROTO_TCP)
+		  { /* send TCP reset */
+	            send_tcp_reset(rif, ip, tcp);
+	            returncode = FW_BLOCK; /* block packet */
+                  }
+		else /* simple reject */
+		  returncode = FW_BLOCK;
+	      }
+	  }
+        else if (rule->fw_rc == SF_RC_ECHO) 
+          {
+            if ((ip->protocol == IPPROTO_ICMP) && (icmp->type == ICMP_ECHO))
+              icmp_reflect(rif, ip, icmp);
+            returncode = FW_BLOCK; /* in any case, block the packet */
+          }
+	else if (rule->fw_rc < FW_REJECT) /* Filter ICMP options */
+	  if (opt == SF_STATE_TRANSMIT) 
+	    returncode = FW_REJECT;
+	  else
+	    {
+	      /* Linux FW functions are limited -- do it yourself... */
+	      switch (rule->fw_rc)
+	        { 
+	          case SF_RC_RNET:   send_icmp_error(rif, ip, ICMP_DEST_UNREACH,
+					       ICMP_NET_UNREACH);
+		  break;
+		  case SF_RC_RPROTO: send_icmp_error(rif, ip, ICMP_DEST_UNREACH,
+					       ICMP_PROT_UNREACH);
+	          break;
+		  case SF_RC_RPORT:  send_icmp_error(rif, ip, ICMP_DEST_UNREACH,
+					       ICMP_PORT_UNREACH);
+		  break;
+	        }
+	      returncode = FW_BLOCK;
+	    }
+
+	if ( !(rule->fw_flags & SF_FW_LOG) ) goto creturn; /* no notification */
+      
+informlog:
 	/* send packet to daemon, append to message buffer */
 	if (sf_log_data.num_entries < NUM_PROC_ENTRIES) {
-	  struct sf_proc_entry *current = &(sf_log_data.data[sf_log_data.num_entries]);
-
-	  current->action   = rule->fw_rc;
-	  current->rule.num = rule->rule_id.num;
-	  strncpy(current->devname,rif->name,7);
-	  current->devname[7]=0;
-	  memcpy(current->ip_packet, ip, MIN(PROC_SIZE_IP, ntohs(ip->tot_len)));
-
+		struct sf_proc_entry *current_p
+				= &(sf_log_data.data[sf_log_data.num_entries]);
+		
+	  current_p->action   = rule->fw_rc;
+	  current_p->rule.num = rule->rule_id.num;
+	  strncpy(current_p->devname,rif->name,7);
+	  current_p->devname[7]=0;
+	  memcpy(current_p->ip_packet, ip, MIN(PROC_SIZE_IP, ntohs(ip->tot_len)));	  
 	  sf_log_data.num_entries++;
 	}
 	else {
@@ -976,14 +1076,15 @@ int sf_check_packet(struct iphdr *ip, struct device *rif, int opt)
 	  wake_up_interruptible(&sf_log_entry);
 
 	  /* drop packet silently if buffer is full */
-	  return SF_RC_BLOCK;
+	  return FW_BLOCK;
 	}
 
 	/* wake up daemon */
 	wake_up_interruptible(&sf_log_entry);
 
+creturn:
 	/* return rule dependent return code */
-	return rule->fw_rc;
+	return returncode;
 }
 
 /* 
@@ -1144,7 +1245,7 @@ int sf_flush(void)
           for (rule=rules; rule!=NULL; rule=rule->fw_next)
             { 
 	      if (   (rule->protocol != IPPROTO_TCP) 
-		  && (rule->fw_flags & SF_FW_PROT != SF_FW_PROT_ALL) )continue;
+		  && ((rule->fw_flags & SF_FW_PROT) != SF_FW_PROT_ALL) )continue;
 	      if (rule->fw_src_cnt)
                 { int match = sf_addr_match(rule->fw_src_idx,rule->fw_src_cnt,
                     sftcp->client_addr,source,rule->fw_flags & SF_FW_SRC_NEG);
@@ -1159,7 +1260,7 @@ int sf_flush(void)
                 }
 	      break; /* rule matches */
             }
-	  if ( (rule==NULL) || (rule->fw_rc!=SF_RC_ACCEPT) )
+	  if ( (rule==NULL) || (rule->fw_rc!=FW_ACCEPT) )
 	    { struct sf_tcp_connection *sftcp2;
               if (sftcp->timer_set) del_timer(&sftcp->timer);
 	      sftcp2 = sftcp->next;
@@ -1237,7 +1338,7 @@ void sf_del_all_timers(void)
   restore_flags(flags);
 }
 
-int sf_write_config(char *buffer, int len)
+int sf_write_config(const char *buffer, int len)
 {
 	struct sf_control *control;
 	struct sf_fw *rule;
@@ -1360,3 +1461,43 @@ int sf_write_config(char *buffer, int len)
 	return rc;
 
 }
+
+
+int sf_forward_chk(struct firewall_ops *this, 
+				   int pf, 
+				   struct device *dev,
+				   void *phdr,
+                   void *arg)
+{
+	return sf_fw_chk(phdr,dev,SF_STATE_FORWARD); 
+}
+
+
+int sf_input_chk(struct firewall_ops *this, 
+				 int pf,
+				 struct device *dev,
+				 void *phdr,
+                 void *arg)
+{
+	return sf_fw_chk(phdr, dev, SF_STATE_RECEIVE); 
+}
+
+
+int sf_output_chk(struct firewall_ops *this, 
+				 int pf,
+				 struct device *dev,
+				 void *phdr,
+                 void *arg)
+{
+	return sf_fw_chk(phdr,dev,SF_STATE_TRANSMIT); 
+}
+
+struct firewall_ops sf_fw_ops = {
+	NULL,
+	sf_forward_chk,
+	sf_input_chk,
+	sf_output_chk,
+	PF_INET,
+	1               /* higher priority than std. firewall */ 
+}; 
+
