@@ -30,7 +30,9 @@
 #define _SF_PROTO_H
 
 extern char *protocol[];
+extern char *icmptype[];
 
-#define MAX_PROTOCOL 98   /* Last non-template entry in protocol array */
+#define MAX_PROTOCOL 103   /* Last non-template entry in protocol array */
+#define MAX_ICMPTYPE  40   /* Last entry in icmptype array */
 
 #endif
