@@ -36,6 +36,9 @@
 #include <unistd.h>
 #include <arpa/inet.h>
 #include <linux/netdevice.h>
+#include <linux/etherdevice.h>
+#include <linux/skbuff.h>
+#include <linux/firewall.h>
 #include <netinet/in.h>
 #include <netinet/ip.h>
 #include <netinet/tcp.h>
@@ -159,7 +162,8 @@ int main(int argc, char **argv)
       show_syntax();
     action = A_STOP;
   }
-  else if (strcmp("reconfig", argv[1]) == 0) {
+  else if ((strcmp("reconfig", argv[1]) == 0) 
+        || (strcmp("restart",  argv[1]) == 0)) {
     action = A_STOP | A_START;
     if (argc > 4)
       show_syntax();
@@ -197,7 +201,8 @@ int main(int argc, char **argv)
       show_syntax();
     action = A_SHOW;
   }
-  else if ((strcmp("checkconfig", argv[1]) == 0) || (strcmp("check", argv[1]) == 0)) {
+  else if ((strcmp("checkconfig", argv[1]) == 0) 
+        || (strcmp("check", argv[1]) == 0)) {
     if (argc > 3)
       show_syntax();
     else if (argc == 3)
@@ -248,7 +253,15 @@ int main(int argc, char **argv)
     close(log);
 
     if ((log = open(SPY_LOG, O_CREAT|O_APPEND|O_RDWR, 0600)) < 0) {
-      perror("Cannot open log file");
+      perror("Cannot open spy log file");
+      exit(1);
+    }
+    if ((getuid() == 0) || (geteuid() == 0))
+      fchown(log, pwd->pw_uid, pwd->pw_gid);
+    close(log);
+
+    if ((log = open(REPORT, O_CREAT|O_APPEND|O_RDWR, 0600)) < 0) {
+      perror("Cannot open report file");
       exit(1);
     }
     if ((getuid() == 0) || (geteuid() == 0))
@@ -343,7 +356,7 @@ int running(void)
 
   /* open the file and lock it */
   if ((fd = open(PID_FILE, O_RDWR)) < 0) {
-    fprintf(stderr, "Cannot open the log file %s. sfc cannot continue.\n", PID_FILE);
+    fprintf(stderr, "Cannot open the file %s. sfc cannot continue.\n", PID_FILE);
     exit(1);
   }
 
@@ -359,7 +372,7 @@ int running(void)
   
   /* read pid from file */
   if ((fdf = fopen(PID_FILE, "r")) == NULL) {
-    fprintf(stderr, "Cannot open the log file %s. sfc cannot continue.\n", PID_FILE);
+    fprintf(stderr, "Cannot open the file %s. sfc cannot continue.\n", PID_FILE);
     exit(1);
   }
 
@@ -578,13 +591,17 @@ static void showcfg(void)
 
       printf("\n    ");
       switch (krule.fw_rc) {
-        case SF_RC_ACCEPT : printf("accept"); 
+        case FW_ACCEPT    : printf("accept"); 
         break;
-        case SF_RC_BLOCK  : printf("block"); 
+	case SF_RC_TREJECT: printf("reject (tcp reset/best)");
+	break;
+	case SF_RC_ECHO   : printf("reject (echo reply)");
+        break;
+        case FW_BLOCK     : printf("block"); 
         break;
         case SF_RC_RNET   : printf("reject (net unreachable)");
         break;
-        case SF_RC_RHOST  : printf("reject (host unreachable)"); 
+        case FW_REJECT    : printf("reject (host unreachable)"); 
         break;
         case SF_RC_RPROTO : printf("reject (protocol unreachable)"); 
         break;
@@ -777,7 +794,11 @@ static void print_addr(unsigned long index, unsigned long count,
       printf(s);
     if (sf_addr[k].mask != SF_IGNORE_ADDRESS) {
       inadr.s_addr = sf_addr[k].addr; printf(inet_ntoa(inadr));
-      inadr.s_addr = sf_addr[k].mask; printf(" mask %s ", inet_ntoa(inadr));
+      if (sf_addr[k].mask != 0xffffffff) {
+        inadr.s_addr = sf_addr[k].mask; printf(" mask %s ", inet_ntoa(inadr));
+      }
+      else
+        printf(" "); 
     }
     if (tcp_or_udp && (sf_addr[k].port != SF_ALL_PORTS)) {
       printf("port %hu", sf_addr[k].port);
